@@ -1,6 +1,6 @@
 /*  cdrdao - write audio CD-Rs in disc-at-once mode
  *
- *  Copyright (C) 1998-2001  Andreas Mueller <mueller@daneb.ping.de>
+ *  Copyright (C) 1998-2002 Andreas Mueller <andreas@daneb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -32,11 +32,32 @@
 #include "PWSubChannel96.h"
 #include "CdTextEncoder.h"
 
+
+
+// Variants for cue sheet generation
+
+// do not set ctl flags for ISRC cue sheet entries
+#define CUE_VAR_ISRC_NO_CTL 0x1
+
+// do not set start of lead-in time in lead-in cue sheet entry
+#define CUE_VAR_CDTEXT_NO_TIME 0x2
+
+#define CUE_VAR_MAX 0x3 
+
+
+// Variants for write parameters mode page
+
+//do not set data block type to block size 2448 if a CD-TEXT lead-in is written
+#define WMP_VAR_CDTEXT_NO_DATA_BLOCK_TYPE 0x1
+
+#define WMP_VAR_MAX 0x1
+
+
 GenericMMC::GenericMMC(ScsiIf *scsiIf, unsigned long options)
   : CdrDriver(scsiIf, options)
 {
   int i;
-  driverName_ = "Generic SCSI-3/MMC - Version 1.2";
+  driverName_ = "Generic SCSI-3/MMC - Version 2.0";
   
   speed_ = 0;
   simulate_ = 1;
@@ -46,6 +67,8 @@ GenericMMC::GenericMMC(ScsiIf *scsiIf, unsigned long options)
   
   cdTextEncoder_ = NULL;
 
+  driveInfo_ = NULL;
+
   memset(&diskInfo_, 0, sizeof(DiskInfo));
 
   for (i = 0; i < maxScannedSubChannels_; i++) {
@@ -70,6 +93,9 @@ GenericMMC::~GenericMMC()
 
   delete cdTextEncoder_;
   cdTextEncoder_ = NULL;
+
+  delete driveInfo_;
+  driveInfo_ = NULL;
 }
 
 // static constructor
@@ -92,6 +118,31 @@ int GenericMMC::checkToc(const Toc *toc)
   return err;
 }
 
+int GenericMMC::subChannelEncodingMode(TrackData::SubChannelMode sm) const
+{
+  int ret = 0;
+
+  switch (sm) {
+  case TrackData::SUBCHAN_NONE:
+    ret = 0;
+    break;
+
+  case TrackData::SUBCHAN_RW:
+    if (options_ & OPT_MMC_NO_RW_PACKED) 
+      ret = 1; // have to encode the R-W sub-channel data
+    else 
+      ret = 0;
+    break;
+
+  case TrackData::SUBCHAN_RW_RAW:
+    // raw R-W sub-channel writing is assumed to be always supported
+    ret = 1;
+    break;
+  }
+
+  return ret;
+}
+
 // sets speed
 // return: 0: OK
 //         1: illegal speed
@@ -107,13 +158,16 @@ int GenericMMC::speed(int s)
 
 int GenericMMC::speed()
 {
-  DriveInfo di;
+  DriveInfo *di;
 
-  if (driveInfo(&di, 1) != 0) {
+  delete driveInfo_;
+  driveInfo_ = NULL;
+
+  if ((di = driveInfo(1)) == NULL) {
     return 0;
   }
 
-  return speed2Mult(di.currentWriteSpeed);
+  return speed2Mult(di->currentWriteSpeed);
 
 }
 
@@ -142,15 +196,62 @@ int GenericMMC::loadUnload(int unload) const
   return 0;
 }
 
+// Checks for ready status of the drive after a write operation
+// Return: 0: drive ready
+//         1: error occured
+//         2: drive not ready
+
+int GenericMMC::checkDriveReady() const
+{
+  unsigned char cmd[10];
+  unsigned char data[4];
+  int ret;
+
+  ret = testUnitReady(0);
+
+  if (ret == 0) {
+    // testUnitReady reports ready status but this might actually not
+    // be the truth -> additionally check the READ DISK INFO command 
+    
+    memset(cmd, 0, 10);
+
+    cmd[0] = 0x51; // READ DISK INFORMATION
+    cmd[8] = 4;
+    
+    ret = sendCmd(cmd, 10, NULL, 0, data, 4, 0);
+    
+    if (ret == 2) {
+      const unsigned char *sense;
+      int senseLen;
+      
+      ret = 0; // indicates ready status
+
+      sense = scsiIf_->getSense(senseLen);
+      
+      if (senseLen >= 14 && (sense[2] & 0x0f) == 0x2 && sense[7] >= 6 &&
+          sense[12] == 0x4 && 
+          (sense[13] == 0x8 || sense[13] == 0x7)) {
+        // Not Ready, long write in progress
+        ret = 2; // indicates not ready status
+      }
+    }
+  }
+
+  return ret;
+}
+
 // Performs complete blanking of a CD-RW.
 // return: 0: OK
 //         1: scsi command failed
 int GenericMMC::blankDisk(BlankingMode mode)
 {
   unsigned char cmd[12];
-  int ret;
+  int ret, progress;
   time_t startTime, endTime;
 
+  setSimulationMode(0);
+
+
   memset(cmd, 0, 12);
 
   cmd[0] = 0xa1; // BLANK
@@ -166,28 +267,41 @@ int GenericMMC::blankDisk(BlankingMode mode)
 
   cmd[1] |= 1 << 4; // immediate return
 
+  sendBlankCdProgressMsg(0);
+
   if (sendCmd(cmd, 12, NULL, 0, NULL, 0, 1) != 0) {
     message(-2, "Cannot erase CD-RW.");
     return 1;
   }
 
   time(&startTime);
+  progress = 0;
 
   do {
     mSleep(2000);
 
-    ret = testUnitReady(0);
+    ret = checkDriveReady();
 
     if (ret == 1) {
       message(-2, "Test Unit Ready command failed.");
     }
+
+    progress += 10;
+    sendBlankCdProgressMsg(progress);
+
+    if (progress >= 1000)
+      progress = 0;
+
   } while (ret == 2);
 
+  if (ret == 0)
+    sendBlankCdProgressMsg(1000);
+
   time(&endTime);
 
   message(2, "Blanking time: %ld seconds", endTime - startTime);
 
-  return 0;
+  return ret;
 }
 
 // sets read/write speed and simulation mode
@@ -336,7 +450,7 @@ int GenericMMC::performPowerCalibration()
 // Sets write parameters via mode page 0x05.
 // return: 0: OK
 //         1: scsi command failed
-int GenericMMC::setWriteParameters()
+int GenericMMC::setWriteParameters(unsigned long variant)
 {
   unsigned char mp[0x38];
   unsigned char mpHeader[8];
@@ -356,8 +470,8 @@ int GenericMMC::setWriteParameters()
     mp[2] |= 1 << 4; // test write
   }
 
-  DriveInfo di;
-  if (driveInfo(&di, 1) == 0 && di.burnProof) {
+  DriveInfo *di;
+  if ((di = driveInfo(1)) != NULL && di->burnProof) {
     // This drive has BURN-Proof function.
     // Enable it unless explicitly disabled.
     if (options_ & OPT_MMC_NO_BURNPROOF) {
@@ -383,10 +497,11 @@ int GenericMMC::setWriteParameters()
                  // used for session at once writing)
 
   if (cdTextEncoder_ != NULL) {
-    mp[4] |= 3; /* Data Block Type: raw data with raw P-W sub-channel data,
-		   block size 2448, required for CD-TEXT lead-in writing
-		   according to the SCSI/MMC-3 manual
-		*/
+    if ((variant & WMP_VAR_CDTEXT_NO_DATA_BLOCK_TYPE) == 0)
+      mp[4] |= 3; /* Data Block Type: raw data with raw P-W sub-channel data,
+		     block size 2448, required for CD-TEXT lead-in writing
+		     according to the SCSI/MMC-3 manual
+		  */
   }
 
   message(4, "Data block type: %u",  mp[4] & 0x0f);
@@ -406,7 +521,7 @@ int GenericMMC::setWriteParameters()
 
   message(4, "Toc type: 0x%x", mp[8]);
 
-  if (setModePage(mp, mpHeader, NULL, 1) != 0) {
+  if (setModePage(mp, mpHeader, NULL, 0) != 0) {
     message(-2, "Cannot set write parameters mode page.");
     return 1;
   }
@@ -414,6 +529,37 @@ int GenericMMC::setWriteParameters()
   return 0;
 }
 
+// Sets simulation mode via mode page 0x05.
+// return: 0: OK
+//         1: scsi command failed
+int GenericMMC::setSimulationMode(int showMessage)
+{
+  unsigned char mp[0x38];
+  unsigned char mpHeader[8];
+
+  if (getModePage(5/*write parameters mode page*/, mp, 0x38,
+		  mpHeader, NULL, showMessage) != 0) {
+    if (showMessage)
+      message(-2, "Cannot retrieve write parameters mode page.");
+    return 1;
+  }
+
+  mp[0] &= 0x7f; // clear PS flag
+
+  if (simulate_)
+    mp[2] |= 1 << 4; // test write
+  else
+    mp[2] &= ~(1 << 4);
+
+  if (setModePage(mp, mpHeader, NULL, showMessage) != 0) {
+    if (showMessage)
+      message(-2, "Cannot set write parameters mode page.");
+    return 1;
+  }
+
+  return 0;
+}
+
 int GenericMMC::getNWA(long *nwa)
 {
   unsigned char cmd[10];
@@ -519,12 +665,34 @@ static unsigned char leadInOutDataMode(TrackData::Mode mode)
   return ret;
 }
 
+
+unsigned char GenericMMC::subChannelDataForm(TrackData::SubChannelMode sm,
+					     int encodingMode)
+{
+  unsigned char ret = 0;
+
+  switch (sm) {
+  case TrackData::SUBCHAN_NONE:
+    break;
+
+  case TrackData::SUBCHAN_RW:
+  case TrackData::SUBCHAN_RW_RAW:
+    if (encodingMode == 0)
+      ret = 0xc0;
+    else
+      ret = 0x40;
+    break;
+  }
+
+  return ret;
+}
 		  
 
 // Creates cue sheet for current toc.
 // cueSheetLen: filled with length of cue sheet in bytes
 // return: newly allocated cue sheet buffer or 'NULL' on error
-unsigned char *GenericMMC::createCueSheet(long *cueSheetLen)
+unsigned char *GenericMMC::createCueSheet(unsigned long variant,
+					  long *cueSheetLen)
 {
   const Track *t;
   int trackNr;
@@ -606,7 +774,8 @@ unsigned char *GenericMMC::createCueSheet(long *cueSheetLen)
 
   cueSheet[n*8+4] = 0;    // Serial Copy Management System
 
-  if (cdTextEncoder_ != NULL) {
+  if (cdTextEncoder_ != NULL &&
+      (variant & CUE_VAR_CDTEXT_NO_TIME) == 0) {
     cueSheet[n*8+5] = leadInStart_.min();
     cueSheet[n*8+6] = leadInStart_.sec();
     cueSheet[n*8+7] = leadInStart_.frac();
@@ -652,6 +821,10 @@ unsigned char *GenericMMC::createCueSheet(long *cueSheetLen)
       }
     }
 
+    // add mode for sub-channel writing
+    dataMode |= subChannelDataForm(t->subChannelType(),
+				   subChannelEncodingMode(t->subChannelType()));
+
     ctl = 0;
     if (t->copyPermitted()) {
       ctl |= 0x20;
@@ -667,7 +840,11 @@ unsigned char *GenericMMC::createCueSheet(long *cueSheetLen)
       }
 
       if (t->isrcValid()) {
-	cueSheet[n*8] = ctl | 0x03;
+	if ((variant & CUE_VAR_ISRC_NO_CTL) == 0)
+	  cueSheet[n*8] = ctl | 0x03;
+	else
+	  cueSheet[n*8] = 0x03;
+
 	cueSheet[n*8+1] = trackNr;
 	cueSheet[n*8+2] = t->isrcCountry(0);
 	cueSheet[n*8+3] = t->isrcCountry(1);
@@ -677,7 +854,11 @@ unsigned char *GenericMMC::createCueSheet(long *cueSheetLen)
 	cueSheet[n*8+7] = t->isrcYear(0) + '0';
 	n++;
 
-	cueSheet[n*8] = ctl | 0x03;
+	if ((variant & CUE_VAR_ISRC_NO_CTL) == 0)
+	  cueSheet[n*8] = ctl | 0x03;
+	else
+	  cueSheet[n*8] = 0x03;
+
 	cueSheet[n*8+1] = trackNr;
 	cueSheet[n*8+2] = t->isrcYear(1) + '0';
 	cueSheet[n*8+3] = t->isrcSerial(0) + '0';
@@ -766,7 +947,7 @@ unsigned char *GenericMMC::createCueSheet(long *cueSheetLen)
   cueSheet[n*8+6] = lostart.sec();
   cueSheet[n*8+7] = lostart.frac();
 
-  message(3, "\nCue Sheet:");
+  message(3, "\nCue Sheet (variant %lx):", variant);
   message(3, "CTL/  TNO  INDEX  DATA  SCMS  MIN  SEC  FRAME");
   message(3, "ADR               FORM");
   for (n = 0; n < len; n++) {
@@ -784,34 +965,55 @@ int GenericMMC::sendCueSheet()
 {
   unsigned char cmd[10];
   long cueSheetLen;
-  unsigned char *cueSheet = createCueSheet(&cueSheetLen);
+  unsigned long variant;
+  unsigned char *cueSheet;
+  int cueSheetSent = 0;
 
-  if (cueSheet == NULL) {
-    return 1;
+  for (variant = 0; variant <= CUE_VAR_MAX; variant++) {
+
+    if (cdTextEncoder_ == NULL &&
+	(variant & CUE_VAR_CDTEXT_NO_TIME) != 0) {
+      // skip CD-TEXT variants if no CD-TEXT has to be written
+      continue;
+    }
+
+    cueSheet = createCueSheet(variant, &cueSheetLen);
+
+    if (cueSheet != NULL) {
+      memset(cmd, 0, 10);
+
+      cmd[0] = 0x5d; // SEND CUE SHEET
+
+      cmd[6] = cueSheetLen >> 16;
+      cmd[7] = cueSheetLen >> 8;
+      cmd[8] = cueSheetLen;
+
+      if (sendCmd(cmd, 10, cueSheet, cueSheetLen, NULL, 0, 0) != 0) {
+	delete[] cueSheet;
+      }
+      else {
+	message(3, "Drive accepted cue sheet variant %lx.", variant);
+	delete[] cueSheet;
+	cueSheetSent = 1;
+	break;
+      }
+    }
   }
 
-  memset(cmd, 0, 10);
-
-  cmd[0] = 0x5d; // SEND CUE SHEET
-
-  cmd[6] = cueSheetLen >> 16;
-  cmd[7] = cueSheetLen >> 8;
-  cmd[8] = cueSheetLen;
-
-  if (sendCmd(cmd, 10, cueSheet, cueSheetLen, NULL, 0) != 0) {
-    message(-2, "Cannot send cue sheet.");
-    delete[] cueSheet;
+  if (cueSheetSent) {
+    return 0;
+  }
+  else {
+    message(-2,
+	    "Drive does not accept any cue sheet variant - please report.");
     return 1;
   }
-
-  delete[] cueSheet;
-  return 0;
 }
 
 int GenericMMC::initDao(const Toc *toc)
 {
   long n;
-  blockLength_ = AUDIO_BLOCK_LEN;
+  blockLength_ = AUDIO_BLOCK_LEN + MAX_SUBCHANNEL_LEN;
   blocksPerWrite_ = scsiIf_->maxDataLen() / blockLength_;
 
   assert(blocksPerWrite_ > 0);
@@ -863,10 +1065,31 @@ int GenericMMC::initDao(const Toc *toc)
 
 int GenericMMC::startDao()
 {
+  unsigned long variant;
+  int writeParametersSet = 0;
+
   scsiTimeout_ = scsiIf_->timeout(3 * 60);
 
-  if (setWriteParameters() != 0)
+  for (variant = 0; variant <= WMP_VAR_MAX; variant++) {
+    if (cdTextEncoder_ == NULL &&
+	(variant & WMP_VAR_CDTEXT_NO_DATA_BLOCK_TYPE) != 0) {
+      // skip CD-TEXT variants if no CD-TEXT has to be written
+      continue;
+    }
+
+    if (setWriteParameters(variant) == 0) {
+      message(3, "Drive accepted write parameter mode page variant %lx.",
+	      variant);
+      writeParametersSet = 1;
+      break;
+    }
+  }
+
+  if (!writeParametersSet) {
+    message(-2, "Cannot setup write parameters for session-at-once mode.");
+    message(-2, "Please try to use the 'generic-mmc-raw' driver.");
     return 1;
+  }
 
   if (!simulate_) {
     if (performPowerCalibration() != 0) {
@@ -895,9 +1118,20 @@ int GenericMMC::startDao()
     return 1;
   }
 
+
   long lba = diskInfo_.thisSessionLba - 150;
 
-  if (writeZeros(toc_->leadInMode(), lba, lba + 150, 150) != 0) {
+  TrackData::Mode mode = TrackData::AUDIO;
+  TrackData::SubChannelMode subChanMode = TrackData::SUBCHAN_NONE;
+  TrackIterator itr(toc_);
+  const Track *tr;
+
+  if ((tr = itr.first()) != NULL) {
+    mode = tr->type();
+    subChanMode = tr->subChannelType();
+  }
+  
+  if (writeZeros(mode, subChanMode, lba, lba + 150, 150) != 0) {
     return 1;
   }
   
@@ -908,10 +1142,13 @@ int GenericMMC::finishDao()
 {
   int ret;
 
-  //message(2, "Writing lead-out...");
+  flushCache(); /* Some drives never return to a ready state after writing
+		 * the lead-out. This is a try to solve this problem.
+		 */
 
-  while ((ret = testUnitReady(0)) == 2)
+  while ((ret = checkDriveReady()) == 2) {
     mSleep(2000);
+  }
 
   if (ret != 0)
     message(-1, "TEST UNIT READY failed after recording.");
@@ -946,13 +1183,13 @@ void GenericMMC::abortDao()
 // by this function.
 // return: 0: OK
 //         1: scsi command failed
-int GenericMMC::writeData(TrackData::Mode mode, long &lba, const char *buf,
-			  long len)
+int GenericMMC::writeData(TrackData::Mode mode, TrackData::SubChannelMode sm,
+			  long &lba, const char *buf, long len)
 {
   assert(blocksPerWrite_ > 0);
   int writeLen = 0;
   unsigned char cmd[10];
-  long blockLength = blockSize(mode);
+  long blockLength = blockSize(mode, sm);
   int retry;
   int ret;
 
@@ -1375,20 +1612,23 @@ int GenericMMC::analyzeTrack(TrackData::Mode mode, int trackNr, long startLba,
 			     char *isrcCode, unsigned char *ctl)
 {
   int ret;
+  int noScan = 0;
 
   selectSpeed(0);
 
-  if (options_ & OPT_MMC_NO_SUBCHAN)
+  if ((readCapabilities_ & CDR_AUDIO_SCAN_CAP) == 0) {
     ret = analyzeTrackSearch(mode, trackNr, startLba, endLba, indexIncrements,
 			     indexIncrementCnt, pregap, isrcCode, ctl);
-  else
+    noScan = 1;
+  }
+  else {
     ret = analyzeTrackScan(mode, trackNr, startLba, endLba,
 			   indexIncrements, indexIncrementCnt, pregap,
 			   isrcCode, ctl);
+  }
 
-  if ((options_ & OPT_MMC_NO_SUBCHAN) ||
-      (options_ & OPT_MMC_READ_ISRC) ||
-      ((options_ & OPT_MMC_USE_PQ) && !(options_ & OPT_MMC_PQ_BCD))) {
+  if (noScan || (options_ & OPT_MMC_READ_ISRC) != 0 ||
+      (readCapabilities_ & CDR_READ_CAP_AUDIO_PQ_HEX) != 0) {
     // The ISRC code is usually not usable if the PQ channel data is
     // converted to hex numbers by the drive. Read them with the
     // appropriate command in this case
@@ -1401,23 +1641,15 @@ int GenericMMC::analyzeTrack(TrackData::Mode mode, int trackNr, long startLba,
   return ret;
 }
 
-int GenericMMC::readSubChannels(long lba, long len, SubChannel ***chans,
+int GenericMMC::readSubChannels(TrackData::SubChannelMode sm,
+				long lba, long len, SubChannel ***chans,
 				Sample *audioData)
 {
   int retries = 5;
   unsigned char cmd[12];
   int i;
   long blockLen;
-
-  if (options_ & OPT_MMC_NO_SUBCHAN) {
-    blockLen = AUDIO_BLOCK_LEN;
-  }
-  else {
-    if (options_ & OPT_MMC_USE_PQ) 
-      blockLen = AUDIO_BLOCK_LEN + 16;
-    else
-      blockLen = AUDIO_BLOCK_LEN + 96;
-  }
+  unsigned long subChanMode = 0;
 
   cmd[0] = 0xbe;  // READ CD
   cmd[1] = 0;
@@ -1429,19 +1661,50 @@ int GenericMMC::readSubChannels(long lba, long len, SubChannel ***chans,
   cmd[7] = len >> 8;
   cmd[8] = len;
   cmd[9] = 0xf8;
-
-  if (options_ & OPT_MMC_NO_SUBCHAN) {
-    cmd[10] = 0;
-  }
-  else {
-    if (options_ & OPT_MMC_USE_PQ) 
-      cmd[10] = 0x02;  // PQ sub-channel data
-    else
-      cmd[10] = 0x01;  // raw PW sub-channel data
-  }
-
   cmd[11] = 0;
 
+  switch (sm) {
+  case TrackData::SUBCHAN_NONE:
+    // no sub-channel data selected choose what is available
+
+    if ((readCapabilities_ & CDR_READ_CAP_AUDIO_PW_RAW) != 0) {
+      // reading of raw P-W sub-channel data is supported
+      blockLen = AUDIO_BLOCK_LEN + 96;
+      cmd[10] = 0x01;  // raw P-W sub-channel data
+
+      subChanMode = CDR_READ_CAP_AUDIO_PW_RAW;
+    }
+    else if ((readCapabilities_ & 
+	      (CDR_READ_CAP_AUDIO_PQ_BCD|CDR_READ_CAP_AUDIO_PQ_HEX)) != 0) {
+      // reading of PQ sub-channel data is supported
+      blockLen = AUDIO_BLOCK_LEN + 16;
+      cmd[10] = 0x02;  // PQ sub-channel data
+
+      if ((readCapabilities_ & CDR_READ_CAP_AUDIO_PQ_BCD) != 0)
+	subChanMode = CDR_READ_CAP_AUDIO_PQ_BCD;
+      else
+	subChanMode = CDR_READ_CAP_AUDIO_PQ_HEX;
+    }
+    else {
+      // no usable sub-channel reading mode is supported
+      blockLen = AUDIO_BLOCK_LEN;
+      cmd[10] = 0;
+
+      subChanMode = 0;
+    }
+    break;
+
+  case TrackData::SUBCHAN_RW:
+    blockLen = AUDIO_BLOCK_LEN + 96;
+    cmd[10] = 0x04;
+    break;
+
+  case TrackData::SUBCHAN_RW_RAW:
+    blockLen = AUDIO_BLOCK_LEN + 96;
+    cmd[10] = 0x01;
+    break;
+  }
+
   while (1) {
     if (sendCmd(cmd, 12, NULL, 0,
 		transferBuffer_, len * blockLen, retries == 0 ? 1 : 0) != 0) {
@@ -1465,37 +1728,56 @@ int GenericMMC::readSubChannels(long lba, long len, SubChannel ***chans,
   }
 #endif
 
-  if ((options_ & OPT_MMC_NO_SUBCHAN) == 0) {
+  if (subChanMode != 0) {
     unsigned char *buf = transferBuffer_ + AUDIO_BLOCK_LEN;
 
     for (i = 0; i < len; i++) {
-      if (options_ & OPT_MMC_USE_PQ) {
-	if (!(options_ & OPT_MMC_PQ_BCD)) {
-	  // All numbers in sub-channel data are hex conforming to the
-	  // MMC standard. We have to convert them back to BCD for the
-	  // 'SubChannel' class.
-	  buf[1] = SubChannel::bcd(buf[1]);
-	  buf[2] = SubChannel::bcd(buf[2]);
-	  buf[3] = SubChannel::bcd(buf[3]);
-	  buf[4] = SubChannel::bcd(buf[4]);
-	  buf[5] = SubChannel::bcd(buf[5]);
-	  buf[6] = SubChannel::bcd(buf[6]);
-	  buf[7] = SubChannel::bcd(buf[7]);
-	  buf[8] = SubChannel::bcd(buf[8]);
-	  buf[9] = SubChannel::bcd(buf[9]);
-	}
-	
+      switch (subChanMode) {
+      case CDR_READ_CAP_AUDIO_PQ_HEX:
+	// All numbers in sub-channel data are hex conforming to the
+	// MMC standard. We have to convert them back to BCD for the
+	// 'SubChannel' class.
+	buf[1] = SubChannel::bcd(buf[1]);
+	buf[2] = SubChannel::bcd(buf[2]);
+	buf[3] = SubChannel::bcd(buf[3]);
+	buf[4] = SubChannel::bcd(buf[4]);
+	buf[5] = SubChannel::bcd(buf[5]);
+	buf[6] = SubChannel::bcd(buf[6]);
+	buf[7] = SubChannel::bcd(buf[7]);
+	buf[8] = SubChannel::bcd(buf[8]);
+	buf[9] = SubChannel::bcd(buf[9]);
+	// fall through
+
+      case CDR_READ_CAP_AUDIO_PQ_BCD:
 	((PQSubChannel16*)scannedSubChannels_[i])->init(buf);
 	if (scannedSubChannels_[i]->type() != SubChannel::QMODE_ILLEGAL) {
 	  // the CRC of the sub-channel data is usually invalid -> mark the
 	  // sub-channel object that it should not try to verify the CRC
 	  scannedSubChannels_[i]->crcInvalid();
 	}
-      }
-      else {
-	((PWSubChannel96*)scannedSubChannels_[i])->init(buf);
-      }
+	break;
       
+      case CDR_READ_CAP_AUDIO_PW_RAW:
+	((PWSubChannel96*)scannedSubChannels_[i])->init(buf);
+	break;
+      }
+
+#if 0
+      if (subChanMode == CDR_READ_CAP_AUDIO_PW_RAW) {
+	// xxam!
+	int j, k;
+	
+	message(0, "");
+	for (j = 0; j < 4; j++) {
+	  for (k = 0; k < 24; k++) {
+	    unsigned char data = buf[j * 24 + k];
+	    message(0, "%02x ", data&0x3f);
+	  }
+	  message(0, "");
+	}
+      }
+#endif
+
       buf += blockLen;
     }
   }
@@ -1506,12 +1788,24 @@ int GenericMMC::readSubChannels(long lba, long len, SubChannel ***chans,
     for (i = 0; i < len; i++) {
       memcpy(audioData, p, AUDIO_BLOCK_LEN);
 
-      p += blockLen;
       audioData += SAMPLES_PER_BLOCK;
+
+      switch (sm) {
+      case TrackData::SUBCHAN_NONE:
+	break;
+
+      case TrackData::SUBCHAN_RW:
+      case TrackData::SUBCHAN_RW_RAW:
+	memcpy(audioData, p + AUDIO_BLOCK_LEN, PW_SUBCHANNEL_LEN);
+	audioData += PW_SUBCHANNEL_LEN / SAMPLE_LEN;
+	break;
+      }
+
+      p += blockLen;
     }
   }
 
-  if (options_ & OPT_MMC_NO_SUBCHAN)
+  if (subChanMode == 0)
     *chans = NULL;
   else
     *chans = scannedSubChannels_;
@@ -1597,25 +1891,32 @@ int GenericMMC::getFeature(unsigned int feature, unsigned char *buf,
   return 0;
 }
 
-int GenericMMC::driveInfo(DriveInfo *info, int showErrorMsg)
+DriveInfo *GenericMMC::driveInfo(int showErrorMsg)
 {
   unsigned char mp[32];
 
+  if (driveInfo_ != NULL)
+    return driveInfo_;
+
+  driveInfo_ = new DriveInfo;
+
   if (getModePage(0x2a, mp, 32, NULL, NULL, showErrorMsg) != 0) {
     if (showErrorMsg) {
       message(-2, "Cannot retrieve drive capabilities mode page.");
     }
-    return 1;
+    delete driveInfo_;
+    driveInfo_ = NULL;
+    return NULL;
   }
 
-  info->burnProof = (mp[4] & 0x80) ? 1 : 0;
-  info->accurateAudioStream = mp[5] & 0x02 ? 1 : 0;
+  driveInfo_->burnProof = (mp[4] & 0x80) ? 1 : 0;
+  driveInfo_->accurateAudioStream = mp[5] & 0x02 ? 1 : 0;
 
-  info->maxReadSpeed = (mp[8] << 8) | mp[9];
-  info->currentReadSpeed = (mp[14] << 8) | mp[15];
+  driveInfo_->maxReadSpeed = (mp[8] << 8) | mp[9];
+  driveInfo_->currentReadSpeed = (mp[14] << 8) | mp[15];
 
-  info->maxWriteSpeed = (mp[18] << 8) | mp[19];
-  info->currentWriteSpeed = (mp[20] << 8) | mp[21];
+  driveInfo_->maxWriteSpeed = (mp[18] << 8) | mp[19];
+  driveInfo_->currentWriteSpeed = (mp[20] << 8) | mp[21];
 
 #if 0
   unsigned char cdMasteringFeature[8];
@@ -1628,7 +1929,7 @@ int GenericMMC::driveInfo(DriveInfo *info, int showErrorMsg)
   }
 #endif
 
-  return 0;
+  return driveInfo_;
 }
 
 TrackData::Mode GenericMMC::getTrackMode(int, long trackStartLba)
@@ -1732,17 +2033,15 @@ CdRawToc *GenericMMC::getRawToc(int sessionNr, int *len)
   return rawToc;
 }
 
-long GenericMMC::readTrackData(TrackData::Mode mode, long lba, long len,
-			       unsigned char *buf)
+long GenericMMC::readTrackData(TrackData::Mode mode,
+			       TrackData::SubChannelMode sm,
+			       long lba, long len, unsigned char *buf)
 {
   long i;
   long inBlockLen = AUDIO_BLOCK_LEN;
   unsigned char cmd[12];
-  TrackData::Mode actMode;
-  int ok = 0;
   const unsigned char *sense;
   int senseLen;
-  int softError;
 
   memset(cmd, 0, 12);
 
@@ -1752,102 +2051,90 @@ long GenericMMC::readTrackData(TrackData::Mode mode, long lba, long len,
   cmd[3] = lba >> 16;
   cmd[4] = lba >> 8;
   cmd[5] = lba;
+  cmd[6] = len >> 16;
+  cmd[7] = len >> 8;
+  cmd[8] = len;
   cmd[9] = 0xf8;
 
-  while (len > 0 && !ok) {
-    cmd[6] = len >> 16;
-    cmd[7] = len >> 8;
-    cmd[8] = len;
+  switch (sm) {
+  case TrackData::SUBCHAN_NONE:
+    cmd[10] = 0; // no sub-channel reading
+    break;
 
-    memset(transferBuffer_, 0, len * inBlockLen);
+  case TrackData::SUBCHAN_RW:
+    cmd[10] = 0x4;
+    inBlockLen += PW_SUBCHANNEL_LEN;
+    break;
 
-    switch (sendCmd(cmd, 12, NULL, 0, transferBuffer_, len * inBlockLen, 0)) {
-    case 0:
-      ok = 1;
-      break;
+  case TrackData::SUBCHAN_RW_RAW:
+    cmd[10] = 0x1;
+    inBlockLen += PW_SUBCHANNEL_LEN;
+    break;
+  }
 
-    case 2:
-      softError = 0;
-      sense = scsiIf_->getSense(senseLen);
+  switch (sendCmd(cmd, 12, NULL, 0, transferBuffer_, len * inBlockLen, 0)) {
+  case 0:
+    break;
 
-      if (senseLen > 0x0c) {
-	if ((sense[2] & 0x0f) == 5) { // Illegal request
-	  switch (sense[12]) {
-	  case 0x63: // End of user area encountered on this track
-	  case 0x64: // Illegal mode for this track
-	    softError = 1;
-	    break;
-	  }
-	}
-	else if ((sense[2] & 0x0f) == 4) { // Hardware Error
-          switch (sense[12]) {
-          case 0x9: // focus servo failure
-            return -2;
-            break;
-          }
-        }    
-	else if ((sense[2] & 0x0f) == 3) { // Medium error
-	  switch (sense[12]) {
-	  case 0x02: // No seek complete, sector not found
-	  case 0x11: // L-EC error
-	  case 0x15: // random positioning error
-	    return -2;
-	    break;
-	  }
+  case 2:
+    sense = scsiIf_->getSense(senseLen);
+
+    if (senseLen > 0x0c) {
+      if ((sense[2] & 0x0f) == 5) { // Illegal request
+	switch (sense[12]) {
+	case 0x63: // End of user area encountered on this track
+	case 0x64: // Illegal mode for this track
+	  return -2;
+	  break;
+
+	case 0x20: // INVALID COMMAND OPERATION CODE
+	case 0x24: // INVALID FIELD IN CDB
+	case 0x26: // INVALID FIELD IN PARAMETER LIST
+	  /* These error codes mean that something was wrong with the
+	   * command we are sending. Report them as hard errors to the
+	   * upper level.
+	   */
+	  scsiIf_->printError();
+	  return -1;
+	  break;
 	}
       }
-
-      if (!softError) {
-	scsiIf_->printError();
-	return -1;
+      else if ((sense[2] & 0x0f) == 4) { // Hardware Error
+	switch (sense[12]) {
+	case 0x9: // focus servo failure
+	  return -2;
+	  break;
+	}
+      }    
+      else if ((sense[2] & 0x0f) == 3) { // Medium error
+	switch (sense[12]) {
+	case 0x02: // No seek complete, sector not found
+	case 0x06: // no reference position found
+	case 0x11: // L-EC error
+	case 0x15: // random positioning error
+	  return -2;
+	  break;
+	}
       }
-      break;
-
-    default:
-      message(-2, "Read error at LBA %ld, len %ld", lba, len);
-      return -1;
-      break;
     }
 
-    if (!ok) {
-      len--;
-    }
+    /* All other errors are unexpected. They will be treated like L-EC errors
+     * by the upper layer. Just print the error code so that we can decice
+     * later to add the errors to the known possible error list.
+     */
+    
+    scsiIf_->printError();
+    return -2;
+    break;
+
+  default:
+    message(-2, "Read error at LBA %ld, len %ld", lba, len);
+    return -2;
+    break;
   }
 
   unsigned char *sector = transferBuffer_;
   for (i = 0; i < len; i++) {
-     
-    if (memcmp(CdrDriver::syncPattern, sector, 12) != 0) {
-      // can't be a data block
-      message(4, "Stopped because no sync pattern found.");
-      message(4, "%d %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x", i, 
-	      sector[0], sector[1], sector[2], sector[3], sector[4], sector[5],
-	      sector[6], sector[7], sector[8], sector[9], sector[10],
-	      sector[11]);
-      return i;
-    }
-    actMode = determineSectorMode(sector + 12);
-
-    if (!(actMode == mode ||
-	  (mode == TrackData::MODE2_FORM_MIX &&
-	   (actMode == TrackData::MODE2_FORM1 ||
-	    actMode == TrackData::MODE2_FORM2)) ||
-
-	  (mode == TrackData::MODE1_RAW && actMode == TrackData::MODE1) ||
-
-	  (mode == TrackData::MODE2_RAW &&
-	   (actMode == TrackData::MODE2 ||
-	    actMode == TrackData::MODE2_FORM1 ||
-	    actMode == TrackData::MODE2_FORM2)))) {
-      message(4, "Sector with not matching mode %s found.",
-	      TrackData::mode2String(actMode));
-      message(4, "%02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x",
-	      sector[12], sector[13], sector[14], sector[15], sector[16],
-	      sector[17], sector[18], sector[19], sector[20], sector[21],
-	      sector[22], sector[23]);
-      return i;
-    }
-
     if (buf != NULL) {
       switch (mode) {
       case TrackData::MODE1:
@@ -1881,8 +2168,34 @@ long GenericMMC::readTrackData(TrackData::Mode mode, long lba, long len,
 	return 0;
 	break;
       }
+
+      // copy sub-channel data
+      switch (sm) {
+      case TrackData::SUBCHAN_NONE:
+	break;
+
+      case TrackData::SUBCHAN_RW:
+      case TrackData::SUBCHAN_RW_RAW:
+	memcpy(buf, sector + AUDIO_BLOCK_LEN, PW_SUBCHANNEL_LEN);
+	buf += PW_SUBCHANNEL_LEN;
+	break;
+      }
     }
 
+#if 0
+    // xxam!
+    int j, k;
+
+    message(0, "");
+    for (j = 0; j < 4; j++) {
+      for (k = 0; k < 24; k++) {
+	unsigned char data = sector[AUDIO_BLOCK_LEN + j * 24 + k];
+	message(0, "%02x ", data&0x3f);
+      }
+      message(0, "");
+    }
+#endif
+
     sector += inBlockLen;
   }
 
@@ -1894,9 +2207,9 @@ int GenericMMC::readAudioRange(ReadDiskInfo *rinfo, int fd, long start,
 			       int endTrack, TrackInfo *info)
 {
   if (!onTheFly_) {
-    if ((options_ & OPT_MMC_NO_SUBCHAN) ||
-	(options_ & OPT_MMC_READ_ISRC) ||
-	((options_ & OPT_MMC_USE_PQ) && !(options_ & OPT_MMC_PQ_BCD))) {
+    if (((readCapabilities_ & CDR_READ_CAP_AUDIO_PQ_BCD) == 0 &&
+	 (readCapabilities_ & CDR_READ_CAP_AUDIO_PW_RAW) == 0) ||
+	(options_ & OPT_MMC_READ_ISRC) != 0) {
       int t;
       long pregap = 0;
 
@@ -1917,7 +2230,7 @@ int GenericMMC::readAudioRange(ReadDiskInfo *rinfo, int fd, long start,
 	sendReadCdProgressMsg(RCD_ANALYZING, rinfo->tracks, t + 1, 0,
 			      totalProgress);
 
-	if (options_ & OPT_MMC_NO_SUBCHAN) {
+	if ((readCapabilities_ & CDR_AUDIO_SCAN_CAP) == 0) {
 	  // we have to use the binary search method to find pre-gap and
 	  // index marks if the drive cannot read sub-channel data
 	  if (!fastTocReading_) {
@@ -1971,8 +2284,14 @@ int GenericMMC::readAudioRange(ReadDiskInfo *rinfo, int fd, long start,
     }
   }
 
-  return CdrDriver::readAudioRangeParanoia(rinfo, fd, start, end, startTrack,
+  if (subChanReadMode_ == TrackData::SUBCHAN_NONE) {
+    return CdrDriver::readAudioRangeParanoia(rinfo, fd, start, end, startTrack,
+					     endTrack, info);
+  }
+  else {
+    return CdrDriver::readAudioRangeStream(rinfo, fd, start, end, startTrack,
 					   endTrack, info);
+  }
 }
 
 int GenericMMC::getTrackIndex(long lba, int *trackNr, int *indexNr,
@@ -2050,3 +2369,307 @@ int GenericMMC::getTrackIndex(long lba, int *trackNr, int *indexNr,
 
   return 0;
 }
+
+/*
+ * Checks if a certain sub-channel reading mode is supported.
+ * lba: start address for reading
+ * len: maximum number of sectors available for testing
+ * subChanMode: 1: read PQ sub-channels
+ *              2: read raw P-W sub-channels
+ *              3: read cooked R-W sub-channels
+ * Return: 0 sub-channel read mode not supported
+ *         1 sub-channel read mode supported (BCD for PQ)
+ *         2 sub-channel read mode supported (HEX for PQ)
+ *         3 sub-channel read mode PQ supported but cannot determine data
+ *           format
+*/
+
+int GenericMMC::readCdTest(long lba, long len, int subChanMode) const
+{
+  unsigned char cmd[12];
+  long blockLen;
+  int ret;
+  int successRead = 0;
+  int pqSubChanBcdOk = 0;
+  int pqSubChanHexOk = 0;
+
+  //message(0, "readCdTest: %ld %ld %d", lba, len, subChanMode);
+
+  if (len <= 0)
+    return 0;
+
+  cmd[0] = 0xbe;  // READ CD
+  cmd[1] = 0;
+  cmd[6] = 0;
+  cmd[7] = 0;
+  cmd[8] = 1; // transfer length: 1
+  cmd[9] = 0xf8;
+  cmd[10] = 0;
+  cmd[11] = 0;
+
+  blockLen = AUDIO_BLOCK_LEN;
+
+  switch (subChanMode) {
+  case 1: // PQ
+    blockLen += PQ_SUBCHANNEL_LEN;
+    cmd[10] = 0x02;
+    if (len > 300)
+      len = 300; /* we have to check many sub-channels here to determine the
+		  * data mode (BCD or HEX)
+		  */
+    break;
+
+  case 2: // PW_RAW
+    cmd[10] = 0x01;
+    blockLen +=  PW_SUBCHANNEL_LEN;
+    if (len > 10)
+      len = 10;
+    break;
+
+  case 3: // RW_COOKED
+    cmd[10] = 0x04;
+    blockLen +=  PW_SUBCHANNEL_LEN;
+    if (len > 10)
+      len = 10;
+    break;
+  }
+
+  while (len > 0) {
+    cmd[2] = lba >> 24;
+    cmd[3] = lba >> 16;
+    cmd[4] = lba >> 8;
+    cmd[5] = lba;
+    
+    if ((ret = sendCmd(cmd, 12, NULL, 0, transferBuffer_, blockLen, 0)) == 0) {
+      successRead++;
+
+      if (subChanMode == 1) {
+	unsigned char *buf = transferBuffer_ + AUDIO_BLOCK_LEN;
+
+#if 0
+	{
+ 	  PQSubChannel16 chan;
+	  chan.init(buf);
+	  chan.print();
+	}
+#endif
+
+	// check if Q sub-channel values are in BCD or HEX format
+	if (SubChannel::isBcd(buf[1]) &&
+	    SubChannel::isBcd(buf[2]) &&
+	    SubChannel::isBcd(buf[3]) &&
+	    SubChannel::isBcd(buf[4]) &&
+	    SubChannel::isBcd(buf[5]) &&
+	    SubChannel::isBcd(buf[6]) &&
+	    SubChannel::isBcd(buf[7]) &&
+	    SubChannel::isBcd(buf[8]) &&
+	    SubChannel::isBcd(buf[9])) {
+	  PQSubChannel16 chan;
+	  chan.init(buf);
+
+	  chan.type(SubChannel::QMODE1DATA);
+
+	  long pqlba =
+	    Msf(chan.amin(), chan.asec(), chan.aframe()).lba() - 150;
+
+	  long diff = pqlba - lba;
+	  if (diff < 0)
+	    diff = -diff;
+	  
+	  if (diff < 20) {
+	    pqSubChanBcdOk++;
+	  }
+	}
+
+	if (buf[7] < 100 && buf[8] < 60 && buf[9] < 75) {
+	  long pqlba = Msf(buf[7], buf[8], buf[9]).lba() - 150;
+
+	  //message(0, "readCdTest: pqlba: %ld", pqlba);
+	  long diff = pqlba - lba;
+	  if (diff < 0)
+	    diff = -diff;
+	  
+	  if (diff < 20) {
+	    pqSubChanHexOk++;
+	  }
+	}
+      }
+    }
+
+
+    len--;
+    lba++;
+  }
+
+  if (successRead) {
+    if (subChanMode == 1) {
+      if (pqSubChanBcdOk > pqSubChanHexOk)
+	return 1;
+      else if (pqSubChanHexOk > pqSubChanBcdOk)
+	return 2;
+      else return 3;
+    }
+    else {
+      return 1;
+    }
+  }
+
+  return 0;
+}
+
+unsigned long GenericMMC::getReadCapabilities(const CdToc *toc,
+					      int nofTracks) const
+{
+  unsigned long caps = 0;
+  int audioRawPWChecked = 0;
+  int audioPQChecked = 0;
+  int audioCookedRWChecked = 0;
+  int dataRawPWChecked = 0;
+  int dataPQChecked = 0;
+  int dataCookedRWChecked = 0;
+  int t;
+
+  if ((options_ & OPT_MMC_NO_SUBCHAN) != 0) {
+    // driver options indicate that PQ and raw RW sub-channel reading for
+    // audio tracks is not supported so skip all corresponding tests
+    audioPQChecked = 1;
+    audioRawPWChecked = 1;
+  }
+  else if ((options_ & OPT_MMC_USE_PQ) != 0) {
+    // driver options indicated that PQ sub-channel reading is supported for
+    // audio/data tracks and RW sub-channel reading is not supported, skip
+    // the corresponding checks and set the capabilities appropriately
+    audioPQChecked = 1;
+    audioRawPWChecked = 1;
+    dataPQChecked = 1;
+    
+    if ((options_ & OPT_MMC_PQ_BCD) != 0)
+      caps |= CDR_READ_CAP_AUDIO_PQ_BCD | CDR_READ_CAP_DATA_PQ_BCD;
+    else
+      caps |= CDR_READ_CAP_AUDIO_PQ_HEX | CDR_READ_CAP_DATA_PQ_HEX;
+  }
+  else if ((options_ & OPT_MMC_USE_RAW_RW) != 0) {
+    // driver options indicated that raw PW sub-channel reading is supported
+    // audio tracks and raw PW sub-channel reading is not supported, skip
+    // the corresponding checks and set the capabilities appropriately
+    audioPQChecked = 1;
+    audioRawPWChecked = 1;
+
+    caps |= CDR_READ_CAP_DATA_PW_RAW;
+  }
+
+  for (t = 0; t < nofTracks; t++) {
+    long tlen = toc[t+1].start - toc[t].start;
+
+    if ((toc[t].adrCtl & 0x04) != 0) {
+      // data track
+      if (!dataPQChecked) {
+	dataPQChecked = 1;
+
+	message(3, "Checking for PQ sub-channel reading support (data track)...");
+	switch (readCdTest(toc[t].start, tlen, 1)) {
+	case 0:
+	  message(3, "PQ sub-channel reading (data track) not supported.");
+	  break;
+
+	case 1:
+	  message(2, "PQ sub-channel reading (data track) is supported, data format is BCD.");
+	  caps |= CDR_READ_CAP_DATA_PQ_BCD;
+	  break;
+	  
+	case 2:
+	  message(2, "PQ sub-channel reading (data track) is supported, data format is HEX.");
+	  caps |= CDR_READ_CAP_DATA_PQ_HEX;
+	  break;
+	  
+	case 3:
+	  message(2, "PQ sub-channel reading (data track) seems to be supported but cannot determine data format.");
+	  message(2, "Please use driver option '--driver generic-mmc:0x1' or '--driver generic-mmc:0x3' to set the data format explicitly.");
+	  break;
+	}
+      }
+	
+      if (!dataRawPWChecked) {
+	dataRawPWChecked = 1;
+
+	message(3, "Checking for raw P-W sub-channel reading support (data track)...");
+	if (readCdTest(toc[t].start, tlen, 2)) {
+	  message(2, "Raw P-W sub-channel reading (data track) is supported.");
+	  caps |= CDR_READ_CAP_DATA_PW_RAW;
+	}
+	else {
+	  message(3, "Raw P-W sub-channel reading (data track) is not supported.");
+	}
+      }
+
+      if (!dataCookedRWChecked) {
+	dataCookedRWChecked = 1;
+      
+	message(3, "Checking for cooked R-W sub-channel reading support (data track)...");
+	if (readCdTest(toc[t].start, tlen, 3)) {
+	  message(2, "Cooked R-W sub-channel reading (data track) is supported.");
+	  caps |= CDR_READ_CAP_DATA_RW_COOKED;
+	}
+	else {
+	  message(3, "Cooked R-W sub-channel reading (data track) is not supported.");
+	}
+      }
+    }
+    else {
+      // audio track
+      if (!audioPQChecked) {
+	audioPQChecked = 1;
+
+	message(3, "Checking for PQ sub-channel reading support (audio track)...");
+	switch (readCdTest(toc[t].start, tlen, 1)) {
+	case 0:
+	  message(3, "PQ sub-channel reading (audio track) is not supported.");
+	  break;
+
+	case 1:
+	  message(2, "PQ sub-channel reading (audio track) is supported, data format is BCD.");
+	  caps |= CDR_READ_CAP_AUDIO_PQ_BCD;
+	  break;
+
+	case 2:
+	  message(2, "PQ sub-channel reading (audio track) is supported, data format is HEX.");
+	  caps |= CDR_READ_CAP_AUDIO_PQ_HEX;
+	  break;
+
+	case 3:
+	  message(2, "PQ sub-channel reading (audio track) seems to be supported but cannot determine data format.");
+	  message(2, "Please use driver option '--driver generic-mmc:0x1' or '--driver generic-mmc:0x3' to set the data format explicitly.");
+	  break;
+	}
+      }
+
+      if (!audioRawPWChecked) {
+	audioRawPWChecked = 1;
+
+	message(3, "Checking for raw P-W sub-channel reading support (audio track)...");
+	if (readCdTest(toc[t].start, tlen, 2)) {
+	  message(2, "Raw P-W sub-channel reading (audio track) is supported.");
+	  caps |= CDR_READ_CAP_AUDIO_PW_RAW;
+	}
+	else {
+	  message(3, "Raw P-W sub-channel reading (audio track) is not supported.");
+	}
+      }
+
+      if (!audioCookedRWChecked) {
+	audioCookedRWChecked = 1;
+
+	message(3, "Checking for cooked R-W sub-channel reading support (audio track)...");
+	if (readCdTest(toc[t].start, tlen, 3)) {
+	  message(2, "Cooked R-W sub-channel reading (audio track) is supported.");
+	  caps |= CDR_READ_CAP_AUDIO_RW_COOKED;
+	}
+	else {
+	  message(3, "Raw R-W sub-channel reading (audio track) is not supported.");
+	}
+      }
+    }
+  }
+
+  return caps;
+}
